from Includes.Defs import *

# MyNode implements the nodes of the Tree.
# The Node has the followign stuff:
# t:    type:   {analog_begin, analog_end, op, if, none, ...}
# v:    value:  the operation or assignment as string
# cnt:  counter:the actual count, how many pathes lead to the this node
# children: depending on the type
# You will find an example in the README
class MyNode:
    # create a node with the given type
    def __init__(self, t=Types.UNKNOWN, v=None, rest=None, cnt=1):

        # check if type is ok:
        if(t == Types.UNKNOWN or type(t) != type(Types.UNKNOWN)):
            print("(!!) Error when adding a node!")
            exit(Errors.WRONG_TYPE)

        # initialize everything
        self._type = t
        self._cnt = cnt

        if (self._type == Types.BEGIN):
            self._child = MyNode(Types.EMPTY, cnt=1)
            return

        if (self._type == Types.END):
            None
            return

        if (self._type == Types.OP):
            self._value = v
            self._child = MyNode(Types.EMPTY)
            return

        if (self._type == Types.BEGIN_IF):
            self._value = v
            self._true = MyNode(Types.BEGIN_TRUE)
            self._false = MyNode(Types.BEGIN_FALSE)
            return

        if (self._type == Types.BEGIN_TRUE):
            self._child = MyNode(Types.EMPTY)
            return

        if (self._type == Types.END_TRUE):
            self._child = MyNode(Types.EMPTY)
            return

        if (self._type == Types.BEGIN_FALSE):
            self._child = MyNode(Types.EMPTY)
            return

        if (self._type == Types.END_FALSE):
            self._child = MyNode(Types.EMPTY)
            return

        if (self._type == Types.END_IF):
            self._child = MyNode(Types.EMPTY)
            return

        if (self._type == Types.SWITCH):
            None
            return

        if (self._type == Types.EMPTY):
            None
            return

# MyTree implements the tree itself. The data
# will be saved in the tree structure. The list-like
# structure was a first idea. If neccessary i will implement
# it later, but for now it just has to work.
# The README has to be updated

# _act_node points to the last node, which usually is a node of type none.
# we need this to implement the branches.

class MyTree:
    _root = None
    _act_node = None
    _stack = []

    # We define the root here.
    def __init__(self):
        self._root = MyNode(Types.BEGIN)
        self._act_node = self._root

    def __str__(self):
        tmp = self._root
        self._stack.clear()
        indentLevel = 0
        while (True):
            prefix = "  " * indentLevel
            if(tmp._type==Types.BEGIN):
                print ("{}{}({})".format(prefix, tmp._type, tmp._cnt))
                tmp = tmp._child
                continue
            if(tmp._type==Types.END):
                print ("{}{}({})".format(prefix, tmp._type, tmp._cnt))
                break
            if(tmp._type==Types.OP):
                print ("{}{}: \"{}\"({})".format(prefix, tmp._type, tmp._value,
                    tmp._cnt))
                tmp = tmp._child
                continue
            if(tmp._type==Types.BEGIN_IF):
                indentLevel+=1
                self._stack.append(tmp._false)
                self._stack.append(tmp._true)
                print ("{}{}: \"{}\"({})".format(prefix, tmp._type, tmp._value,
                    tmp._cnt))
                tmp = self._stack.pop()
                continue
            if(tmp._type==Types.BEGIN_TRUE):
                indentLevel+=1
                print ("{}{}({})".format(prefix, tmp._type, tmp._cnt))
                tmp = tmp._child
                continue
            if(tmp._type==Types.END_TRUE):
                indentLevel-=1
                prefix = "  " * indentLevel
                print ("{}{}({})".format(prefix, tmp._type, tmp._cnt))
                tmp = self._stack.pop()
                continue
            if(tmp._type==Types.BEGIN_FALSE):
                indentLevel+=1
                print ("{}{}({})".format(prefix, tmp._type, tmp._cnt))
                tmp = tmp._child
                continue
            if(tmp._type==Types.END_FALSE):
                indentLevel-=1
                prefix = "  " * indentLevel
                print ("{}{}({})".format(prefix, tmp._type, tmp._cnt))
                tmp = tmp._child
                continue
            if(tmp._type==Types.END_IF):
                indentLevel -=1
                prefix = "  " * indentLevel
                print ("{}{}({})".format(prefix, tmp._type, tmp._cnt))
                tmp = tmp._child
        return ''
    
    # add a node and change self._act_node
    def add_node(self, node=None):
        if (node == None):
            print("(!!) no node was given!")
            exit(Errors.NO_NODE)

        if (self._act_node._type == Types.END):
            print("(!!) Trying to add node after end!")
            exit(Errors.ADD_AFTER_END)
            
        if (node._type == Types.BEGIN_IF):
            # set actual node (including botch children)
            self._act_node._child = node
            self._act_node._child._cnt = self._act_node._cnt
            self._act_node = self._act_node._child
            self._act_node._true._cnt = self._act_node._cnt
            self._act_node._false._cnt = self._act_node._cnt

            # push false and true on stack
            self._stack.append(self._act_node._false)
            self._stack.append(self._act_node._true)
            # and pop true. False is still on the stack
            self._act_node = self._stack.pop()

        if (node._type == Types.END_TRUE):
            self._act_node._child = node
            self._act_node._child._cnt = self._act_node._cnt
            # save the END_TRUE-node as tmp and push on stack
            tmp = self._act_node._child
            self._act_node = self._stack.pop()
            self._stack.append(tmp)

        if (node._type == Types.END_FALSE):
            # change act_node._child to END_FALSE
            self._act_node._child = node
            self._act_node._child._cnt = self._act_node._cnt
            # and push END_FALSE on the stack
            self._stack.append(self._act_node._child)
            # now 2 END_NODES should be on the stack!

        if (node._type == Types.END_IF):
            tmp = self._stack.pop()
            cnt = tmp._cnt
            tmp._child = node

            tmp = self._stack.pop()
            cnt += tmp._cnt
            tmp._child = node
            self._act_node = node
            self._act_node._cnt = cnt


        if (node._type == Types.OP):
            self._act_node._child = node
            self._act_node._child._cnt = self._act_node._cnt
            self._act_node =  self._act_node._child

        if (node._type == Types.END):
            self._act_node._child = node
            self._act_node._child._cnt = self._act_node._cnt
            self._act_node =  self._act_node._child

    def print_tree(self, notUsedVariable=None):
        print(self.__str__())
