import sys
from antlr4 import *

from Includes.Defs import *

from Parser.VerilogAParser import VerilogAParser
from Parser.VerilogAListener import VerilogAListener
from Parser.VerilogALexer import VerilogALexer

from ReadAndPrint.ReadAndPrintTree import *

# the module, holding the built tree
class Module:
    _tree = MyTree()

# the class, building and modifying the tree
class ReadAndPrintListener(VerilogAListener):

    # initialize the tree
    _module = Module()
    # this listener builds the AST while walking along the ParseTree

    def getModule(self):
        return self

    def writeModule(self, FileName=None):
        if FileName == None:
            sys.exit()
        f = open(FileName, mode='w')
        f.write("Module " + self._module.modName + "\n")
        f.close()

    def enterPreprocessor_items(self, ctx):
        if ctx.comment() != None:
            self._module._tree.add_node(MyNode(Types.COMMENT, ctx.getText()[:-1]))
        else:
            self._module._tree.add_node(MyNode(Types.PREPROCESSOR, ctx.getText()[:-1]))

    def enterModule_declaration (self, ctx):
        # check, if ins_mod exits:
        txt = ""
        if ctx.ins_mod() != None:
            txt += ctx.ins_mod().getText()
        txt += "module "
        txt += ctx.module_identifier().getText()
        txt += ctx.list_of_ports().getText()
        txt += ";"
        
        self._module._tree.add_node(MyNode(Types.MODULE_DECLARATION,
            "{}".format(txt)
            ))

    def enterModule_item_declaration(self, ctx):
        if ctx.inout_declaration() != None:
            self._module._tree.add_node(MyNode(Types.DECLARATION, "inout {};".format(ctx.inout_declaration().list_of_port_identifiers().getText())))
        if ctx.input_declaration() != None:
            self._module._tree.add_node(MyNode(Types.DECLARATION, "input {};".format(ctx.input_declaration().list_of_port_identifiers().getText())))
        if ctx.output_declaration() != None:
            self._module._tree.add_node(MyNode(Types.DECLARATION, "output {};".format(ctx.output_declaration().list_of_port_identifiers().getText())))
        if ctx.real_declaration() != None:
            self._module._tree.add_node(MyNode(Types.DECLARATION, "real {};".format(ctx.real_declaration().list_of_identifiers().getText())))
        if ctx.integer_declaration() != None:
            self._module._tree.add_node(MyNode(Types.DECLARATION, "integer {};".format(ctx.integer_declaration().list_of_identifiers().getText())))

    def exitSource_text(self, ctx):
        self._module._tree.add_node(MyNode(Types.END))

    def enterNode_declaration(self, ctx):
        self._module._tree.add_node(MyNode(Types.DECLARATION, "{} {};".format(ctx.discipline_identifier().getText(), ctx.list_of_nodes().getText())))

    def enterParameter_declaration(self, ctx):
        txt = ""
        txt += "parameter {} ".format(ctx.opt_type().getText())
        lopa = ctx.list_of_param_assignments()
        while(lopa != None):
            txt += lopa.declarator_init().parameter_identifier().getText() + " = "
            if lopa.declarator_init().constant_expression() != None:
                txt += lopa.declarator_init().constant_expression().getText() + " " 
            else:
                txt += lopa.declarator_init().define_expression().getText() + " "
            if lopa.declarator_init().opt_range(0) != None:
                txt += lopa.declarator_init().opt_range(0).getText()
            lopa = lopa.list_of_param_assignments()

        if ctx.comment() != None:
            txt += ctx.comment().getText()[:-1]
        txt += ";"
        self._module._tree.add_node(MyNode(Types.DECLARATION, txt))

    def exitModule_declaration(self, ctx):
        self._module._tree.add_node(MyNode(Types.MODULE_DECLARATION, "endmodule"))

    def enterAnalog_block(self, ctx):
        self._module._tree.add_node(MyNode(Types.BEGIN_ANALOG, 'analog'))

    def recExp(self, ctx):
        if ctx.binary_operator() != None:
            return self.recExp(ctx.expression(0)) + " " + ctx.binary_operator().getText() + " " + self.recExp(ctx.expression(1))
        elif ctx.primary() != None and ctx.primary().expression() != None:
            return "(" + self.recExp(ctx.primary().expression()) + ")"
        elif ctx.built_in_function() != None:
            if len(ctx.arg_list().children) == 1:
                return ctx.built_in_function().getText() + "(" + self.recExp(ctx.arg_list().expression(0))+ ")"
            else:
                txt = ""
                txt += ctx.built_in_function().getText() + "("
                txt += self.recExp(ctx.arg_list().expression(0))
                for i in range(2,len(ctx.arg_list().children)+1, 2):
                    txt += ", " + self.recExp(ctx.arg_list().expression(i/2))
                txt += ")"
                return txt
        else:
            return ctx.getText()

    def recEventExpression(self, ctx):
        txt = ""
        if ctx.simple_event().global_event() != None:
            txt += ctx.simple_event().global_event().children[0].getText()
            if ctx.simple_event().global_event().analysis_list() != None:
                txt += " (" + ctx.simple_event().global_event().analysis_list().getText() + " )"

        if ctx.simple_event().event_function() != None:
            print(ctx.simple_event().event_function().getText())
            txt = "timer(" + self.recExp(ctx.simple_event().event_function().timer_function().start_time().expression())
            if ctx.simple_event().event_function().timer_function().period() != None:
                txt += ", " + self.recExp(ctx.simple_event().event_function().timer_function().period().expression())
            txt += ")"

        if ctx.simple_event().identifier() != None:
            print("(!!) Identifier and subexpressions not yet implemented!")
            print("(!!) found at {}".format(ctx.start.line))
            sys.exit(Errors.NOT_YET_IMPLEMENTED)

        if ctx.simple_event().children[0].getText() == "cross(":
            txt += "cross( "
            txt += self.recExp(ctx.simple_event().expression(0))
            if ctx.simple_event().direction() != None:
                txt += ", " + ctx.simple_event().direction().getText()
            if ctx.simple_event().expression(1) != None:
                txt += ", " + self.recExp(ctx.simple_event().expression(1))
            if ctx.simple_event().parent_identifier() != None:
                txt += ", " + ctx.simple_event().parent_identifier().getText()
            if ctx.simple_event().attribute() != None:
                txt += ", " + ctx.simple_event().attribute().getText()
            txt += ")"
            

        if ctx.event_expression() != None:
            txt += " or " + self.recEventExpression(ctx.event_expression())

        return txt

    def enterStatement(self, ctx):
        if ctx.comment() != None:
            self._module._tree.add_node(MyNode(Types.COMMENT, ctx.getText()))
        if ctx.block_statement() != None:
            self._module._tree.add_node(MyNode(Types.OP, "begin"))
        if ctx.branch_contribution() != None:
            self._module._tree.add_node(MyNode(Types.OP, ctx.branch_contribution().bvalue().getText() + " <+ " + self.recExp(ctx.branch_contribution().expression()) + ";"))
        if ctx.procedural_assignment() != None:
            self._module._tree.add_node(MyNode(Types.OP, ctx.procedural_assignment().lexpr().getText() + " = " + self.recExp(ctx.procedural_assignment().expression()) + ";"))
        if ctx.event_controlled_statement() != None:
            self._module._tree.add_node(MyNode(Types.OP, "@ ( " + self.recEventExpression(ctx.event_controlled_statement().event_expression()) + ")"))
        if ctx.simulator_function() != None:
            txt = ""
            two = False
            for i in range(ctx.getChildCount()):
                print(i)
                two = False
                if (ctx.getChild(i).getText() == "$"):
                    two = True
                elif (ctx.getChild(i).getText() == "%"):
                    two = True
                elif (ctx.getChild(i).getText() == "\""):
                    two = True
                elif (i<ctx.getChildCount()-1 and ctx.getChild(i+1).getText() == "\""):
                    two = True
                elif (ctx.getChild(i).getText() == "("):
                    two = True
                elif (i<ctx.getChildCount()-1 and ctx.getChild(i+1).getText() == "("):
                    two = True
                elif (i<ctx.getChildCount()-1 and ctx.getChild(i+1).getText() == ")"):
                    two = True
                if (two):
                    txt += ctx.getChild(i).getText()
                else:
                    txt += ctx.getChild(i).getText()
                    txt += " "
                print(txt)
            txt+=";"
            self._module._tree.add_node(MyNode(Types.OP, txt))
        if ctx.finish_function() != None:
            self._module._tree.add_node(MyNode(Types.OP, ctx.finish_function().getText()+";"))

    def exitStatement(self, ctx):
        if ctx.block_statement() != None:
            self._module._tree.add_node(MyNode(Types.OP, "end"))

    def enterConditional_statement(self, ctx):
        txt = ""
        txt += "if("
        if ctx.expression().binary_operator() != None:
            txt += ctx.expression().expression(0).getText() + " "
            txt += ctx.expression().binary_operator().getText() + " "
            txt += ctx.expression().expression(1).getText()
        else:
            txt += ctx.expression().getText()
        txt += ")"
        self._module._tree.add_node(MyNode(Types.BEGIN_IF, txt))

    def enterFalse_statement(self, ctx):
        self._module._tree.add_node(MyNode(Types.BEGIN_FALSE, "else"))

    def exitConditional_statement(self, ctx):
        if ctx.false_statement() == None:
            self._module._tree.add_node(MyNode(Types.BEGIN_FALSE, """else
            begin
            end"""))
